/*
 * Photo Image Print System
 * Copyright (C) 2000-2005 EPSON AVASYS Corporation.
 * Copyright (C) Seiko Epson Corporation 2000-2008.
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, EPSON AVASYS Corporation gives permission to
 * link the code of this program with libraries which are covered by
 * the EPSON AVASYS Public License and distribute their linked
 * combinations.  You must obey the GNU General Public License in all
 * respects for all of the code used other than the libraries which
 * are covered by EPSON AVASYS Public License.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/stat.h>
#include <libxml/parser.h>
#include "pipscap-util.h"
#include "pipscap-data.h"
#define COMMON_DATA_PATH "/usr/local/EPAva/core/epson-common.xml"

/***************************Define internal function***************************/
static int pipscap_parser_file (PipsCapParserPtr, char *);
static int CheckExistFile (const char *);
static int OverCheckXmlFile (const char *, xmlDocPtr*, char *);
static int RemovePrinterFromPipsrcFile (const char *, char *);
static int CheckExistPrinterFromPipsrcFile (const char *, char *);
/*****************************************************************************/

static void
pipscap_parser_start_document (PipsCapParser *state)
{
    state->state = PARSER_START;

    state->id = NULL;
    state->content = NULL;
    state->prop_name = NULL;

    state->ret_data = NULL;
    state->ret_code = 0;

    state->list_type = LST_NONE;

    if (state->parser_status == 1)
    {
        state->data_list = (CapDatabasePtr) malloc (sizeof (CapDatabase));
        state->data_list->cap_list = NULL;
        state->data_list->other_list = NULL;
        state->data_list->papersize_list = NULL;
        state->data_list->cons_list = NULL;
        state->data_list->mode_color_list = NULL;
        state->data_list->mode_mono_list = NULL;
        state->data_list->mode_abwp_list = NULL;
        state->data_list->name_def_list = NULL;
    }
    else
        state->data_list->common_list = NULL;

}

static void
pipscap_parser_end_document(PipsCapParser *state)
{
    if (state->content != NULL)
    {
    	free (state->content);
    	state->content = NULL;
    }
    
    if (state->prop_name)
    {
    	xmlFree (state->prop_name);
    	state->prop_name = NULL;
    }
    return;
}

static void
pipscap_parser_start_element(PipsCapParser *state,
                             const xmlChar *name, const xmlChar **attrs)
{
    int i = 0;
    int rawfilter = 0;
    char *param = NULL;

    switch (state->state)
    {
        /* Reading specific printer firstly */
    case PARSER_START:
        if (!strcmp(name, "printer"))
        {
            state->state = PARSER_PRINTER;
            if (state->parser_status == 1) /* Specific printer */
            {
                for (i = 0; attrs && attrs[i] != NULL; i += 2)
                {
                    int id = -1;

                    if (!strcmp (attrs[i], "name"))
                    {
                        id = P_PRINTER_NAME;
                    }
                    else if (!strcmp (attrs[i], "model_name"))
                    {
                        id = P_MODEL_NAME;
                    }
                    else if (!strcmp (attrs[i], "filter_name"))
                    {
                        id = P_FILTER_NAME;
                    }

                    if (id < 0)
                    {   /* Error */
                        state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        break;
                    }
                    else if (AddContextListCap (&state->data_list->name_def_list, id , (char *)attrs[i+1]))
                    {   /* Error */
                        state->ret_code = MEMORY_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        break;
                    }
                }
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = (char *)name;
            return;
        }
        break;

    case PARSER_PRINTER:
        if (!strcmp(name, "feature"))
        {
            state->state = PARSER_FEATURE;
            for (i = 0; attrs && attrs[i] != NULL; i += 2)
            {
                if (!strcmp (attrs[i], "id"))
                    state->id = (char *)attrs[i+1];
                else if (!strcmp (attrs[i], "param"))
                    param = (char *)attrs[i+1];
                else  if(!strcmp (attrs[i], "rawfilter"))
                    rawfilter = (int) atoi ((char *)attrs[i+1]);
                else
                {
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    state->ret_data = strdup (attrs[i]);
                    return;
                }
            }
            state->rawfilter = rawfilter;
            if (state->parser_status == 1)
            {
                if (atoi (state->id) == P_OTHER)
                {
                    state->list_type = LST_OTHER;
                }
                else if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
                {
                    if (atoi (state->id) == P_MEDIA_SIZE)
                    {
                        state->list_type = LST_MEDIASIZE;
                        if (AddCapListCap (&state->data_list->cap_list, atoi (state->id), param))
                        {
                            state->ret_code = MEMORY_ERROR;
                            return;
                        }
                    }
                    else
                    {
                        state->list_type = LST_CAP;
                        if (AddCapListCap (&state->data_list->cap_list, atoi (state->id), param))
                        {
                            state->ret_code = MEMORY_ERROR;
                            return;
                        }
                    }
                }
            }
            else /* Update common option list into cap list */
            {
                state->list_type = LST_COMMON;
                if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
                {
                    if (AddCapListCap (&state->data_list->common_list, atoi (state->id), param))
                    {
                        state->ret_code = MEMORY_ERROR;
                        return;
                    }
                }
            }
        }
        else if (!strcmp(name, "constraint"))
        {
            state->state = PARSER_CONSTRAINT;
            for (i = 0; attrs && attrs[i] != NULL; i += 2)
            {
                if (!strcmp (attrs[i], "id"))
                    state->id = (char *) attrs[i+1];
                else if (!strcmp (attrs[i],"rawfilter"))
                    rawfilter = atoi ((char *)attrs[i+1]);
                else
                {
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    state->ret_data = strdup (attrs[i]);
                    return;
                }
            }
            state->rawfilter = rawfilter;

            /* Create new constraint papepsource node */
            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
            {
                state->list_type = LST_CONSTRAINT;
                if (AddConsPaperSourceListCap (&state->data_list->cons_list, state->id))
                {
                    state->ret_code = MEMORY_ERROR;
                    return;
                }
            }
        }
        else if (!strcmp(name, "printmode"))
        {
            state->state = PARSER_PRINTMODE;
            for (i = 0; attrs && attrs[i] != NULL; i += 2)
            {
                if (!strcmp (attrs[i], "name"))
                    state->id = (char *)attrs[i+1];
                else
                {
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    state->ret_data = strdup (attrs[i]);
                    return;
                }
            }
            if (!strcmp (state->id, "COLOR") || !strcmp (state->id, "COLOR/MONO"))
                state->list_type = LST_MODE_COLOR;
            else if (!strcmp (state->id, "MONO"))
                state->list_type = LST_MODE_MONO;
            else if (!strcmp (state->id, "ABWP"))
                state->list_type = LST_MODE_ABWP;
            else
            {
                state->ret_code = RESOURCE_VALUES_ERROR;
                if (attrs[i] != 0)
                {
                    state->ret_data = strdup (attrs[i]);
                }
                return;
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        break;

    case PARSER_FEATURE:
        if (!strcmp(name, "option"))
        {
            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1))
                    ||(state->list_type == LST_OTHER))
            {
                state->state = PARSER_OPTION;
                for (i = 0; attrs && attrs[i] != NULL; i += 2)
                {
                    if (!strcmp (attrs[i], "id"))
                        state->id = (char *) xmlStrdup (attrs[i+1]);
                    else
                    {
                        state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        return;
                    }
                }

                if (state->list_type == LST_MEDIASIZE)
                {   /* Create new papepsize node */
                    if (AddPaperSizeListCap (&state->data_list->papersize_list, state->id))
                    {
                        state->ret_code = MEMORY_ERROR;
                        return;
                    }
                }
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        break;

    case PARSER_CONSTRAINT:
        if (!strcmp(name, "option"))
        {
            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
            {
                state->state = PARSER_OPTION;
                for (i = 0; attrs && attrs[i] != NULL; i += 2)
                {
                    if (!strcmp (attrs[i], "id"))
                        state->id = (char *)attrs[i+1];
                    else
                    {
                        state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        return;
                    }
                }
                /* Add new constrain paper size list */
                if (AddConsPaperSizeListCap (&state->data_list->cons_list->list, state->id))
                {
                    state->ret_code = MEMORY_ERROR;
                    return;
                }
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        break;

    case PARSER_PRINTMODE:
        if (!strcmp(name, "mode"))
        {
            char *mode_name = NULL;
            for (i = 0; attrs && attrs[i] != NULL; i += 2)
            {
                if (!strcmp (attrs[i], "rawfilter"))
                    rawfilter = (int) atoi ((char *)attrs[i+1]);
                else if (!strcmp (attrs[i], "name"))
                    mode_name = (char *)attrs[i+1];
                else if (!strcmp (attrs[i], "default"))
                {
                    if (atoi ((char *)attrs[i+1]) == 1)
                    {
                        int modeid = 3;
                        char str_mode[3];
                        if (!state->data_list->mode_color_list)
                            modeid = 1;
                        else
                            modeid = state->data_list->mode_color_list->modeid + 1;
                        sprintf (str_mode, "%d", modeid);
                        if (AddContextListCap (&state->data_list->name_def_list, P_DEF_MODE, str_mode))
                        {
                            state->ret_code = MEMORY_ERROR;
                            state->ret_data = strdup (attrs[i]);
                            return;
                        }
                    }
                }
                else
                {
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    state->ret_data = strdup (attrs[i]);
                    return;
                }
            }
            state->rawfilter = rawfilter;

            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
            {
                PrintModeList *list;
                state->state = PARSER_MODE;

                switch (state->list_type)
                {
                case LST_MODE_COLOR:
                    list = &state->data_list->mode_color_list;
                    break;
                case LST_MODE_MONO:
                    list = &state->data_list->mode_mono_list;
                    break;
                case LST_MODE_ABWP:
                    list = &state->data_list->mode_abwp_list;
                    break;
                default:
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    return;
                }

                if (AddPrintModeListCap (list, mode_name))
                {
                    state->ret_code = MEMORY_ERROR;
                    return;
                }
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        break;

    case PARSER_MODE:
        if (!strcmp(name, "option"))
        {
            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
            {
                state->state = PARSER_OPTION;
                for (i = 0; attrs && attrs[i] != NULL; i += 2)
                {
                    if (!strcmp (attrs[i], "id"))
                        state->id = (char *) xmlStrdup (attrs[i+1]);
                    else
                    {
                        state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        return;
                    }
                }
            }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        break;

    case PARSER_OPTION:
        if (!strcmp(name, "property"))
        {
            if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
            {
                state->state = PARSER_PROPERTY;
                for (i = 0; attrs && attrs[i] != NULL; i += 2)
                {
                    if (!strcmp (attrs[i], "name"))
                        state->prop_name = (char *)xmlStrdup (attrs[i+1]);
                    else
                    {
                        state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                        state->ret_data = strdup (attrs[i]);
                        return;
                    }
                }
            }

	    if (state->content != NULL)
	    {
		free (state->content);
		state->content = NULL;
	    }
        }
        else
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
        }
        break;

    case PARSER_PROPERTY:
    case PARSER_FINISH:
        state->ret_code = RESOURCE_OPTION_ERROR;
        state->ret_data = strdup (name);
        break;
    }

    return;
}

static void
pipscap_parser_end_element (PipsCapParser *state, const xmlChar *name)
{
    ParserState parser_state = state->state;

    switch (parser_state)
    {
    case PARSER_START:
        /* None */
        break;

    case PARSER_PRINTER:
        if (strcmp (name, "printer"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        state->state = PARSER_FINISH;
        break;

    case PARSER_FEATURE:
        if (strcmp (name, "feature"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }

        state->state = PARSER_PRINTER;
        break;

    case PARSER_CONSTRAINT:
        if (strcmp (name, "constraint"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        state->state = PARSER_PRINTER;
        break;

    case PARSER_PRINTMODE:
        if (strcmp (name, "printmode"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        state->state = PARSER_PRINTER;
        break;

    case PARSER_MODE:
        if (strcmp (name, "mode"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }
        state->state = PARSER_PRINTMODE;
        break;

    case PARSER_OPTION:
        if (strcmp (name, "option"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }

        if (state->list_type == LST_OTHER) /* Common */
        {
            if (AddContextListCap (&state->data_list->other_list, atoi (state->id), state->content))
            {
                state->ret_code = MEMORY_ERROR;
                return;
            }
            state->state = PARSER_FEATURE;
        }
        else if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
        {
            if (state->parser_status == 1)
            {
                int result = 0;

                switch (state->list_type)
                {
                case LST_NONE:
                case LST_OTHER:
                case LST_COMMON:
                    /* Not handled */
                    break;

                case LST_CAP:
                case LST_MEDIASIZE:
                    result = AddElementListCap (&state->data_list->cap_list->list, state->id, _(state->content));
                    state->state = PARSER_FEATURE;
                    break;

                case LST_CONSTRAINT:
                    state->state = PARSER_CONSTRAINT;
                    break;

                case LST_MODE_COLOR:
                    result = AddContextListCap (&state->data_list->mode_color_list->list, (int) atoi (state->id), state->content);
                    state->state = PARSER_MODE;
                    break;

                case LST_MODE_MONO:
                    result = AddContextListCap (&state->data_list->mode_mono_list->list, (int) atoi (state->id), state->content);
                    state->state = PARSER_MODE;
                    break;

                case LST_MODE_ABWP:
                    result = AddContextListCap (&state->data_list->mode_abwp_list->list, (int) atoi (state->id), state->content);
                    state->state = PARSER_MODE;
                    break;

                default:
                    state->ret_code = RESOURCE_ATTRIBUTE_ERROR;
                    state->state = parser_state;
                    return;
                }

                if (result != 0)
                {
                    state->ret_code = MEMORY_ERROR;
                    state->state = parser_state;
                    return;
                }
            }
            else /* Common data */
            {
                if (AddElementListCap (&state->data_list->common_list->list, state->id, _(state->content)))
                {
                    state->ret_code = MEMORY_ERROR;
                    return;
                }
                state->state = PARSER_FEATURE;
            }
        }
        else
        {
            if (state->parser_status == 1)
            {
                if (state->list_type == LST_CAP || state->list_type == LST_MEDIASIZE)
                    state->state = PARSER_FEATURE;
                else if (state->list_type == LST_CONSTRAINT)
                    state->state = PARSER_CONSTRAINT;
                else if ((state->list_type == LST_MODE_COLOR) || (state->list_type == LST_MODE_MONO) || (state->list_type == LST_MODE_ABWP))
                    state->state = PARSER_MODE;
            }
            else /* Common */
                state->state = PARSER_FEATURE;
        }
        break;

    case PARSER_PROPERTY:
        if (strcmp (name, "property"))
        {
            state->ret_code = RESOURCE_OPTION_ERROR;
            state->ret_data = strdup (name);
            return;
        }

        if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1)))
        {
            if (state->list_type == LST_CONSTRAINT)
            {
                if (!strcmp (state->prop_name, "pawidth"))
                    state->data_list->cons_list->list->pa_width = atol (state->content);
                else if (!strcmp (state->prop_name, "palength"))
                    state->data_list->cons_list->list->pa_length = atol (state->content);
                else if (!strcmp (state->prop_name, "palength"))
                    state->data_list->cons_list->list->pa_length = atol (state->content);
                else if (!strcmp (state->prop_name, "leftmargin"))
                    state->data_list->cons_list->list->left_margin = atol (state->content);
                else if (!strcmp (state->prop_name, "rightmargin"))
                    state->data_list->cons_list->list->right_margin = atol (state->content);
                else if (!strcmp (state->prop_name, "topmargin"))
                    state->data_list->cons_list->list->top_margin = atol (state->content);
                else if (!strcmp (state->prop_name, "bottommargin"))
                    state->data_list->cons_list->list->bottom_margin = atol (state->content);
            }
            else if (state->list_type == LST_MEDIASIZE)
            {
                if (!strcmp (state->prop_name, "name"))
                {
                    if (AddElementListCap (&state->data_list->cap_list->list, state->id, _(state->content)))
                    {
                        state->ret_code = MEMORY_ERROR;
                        return;
                    }
                }
                else if (!strcmp (state->prop_name, "width"))
                    state->data_list->papersize_list->width = atol (state->content);
                else if (!strcmp (state->prop_name, "length"))
                    state->data_list->papersize_list->height = atol (state->content);
            }
        }
        state->state = PARSER_OPTION;
        break;

    case PARSER_FINISH:
        /* Finish parser xml file */
        break;
    }

    /* Clean up */
    if (state->content != NULL)
    {
	    free (state->content);
	    state->content = NULL;
    }
    return;
}

static void
pipscap_parser_characters (PipsCapParser *state, const xmlChar *chars, int len)
{
    switch (state->state)
    {
    case PARSER_OPTION:
    case PARSER_PROPERTY:
        if ((state->filter_type == 1) || ((state->filter_type == 2) && (state->rawfilter == 1))
                ||(state->list_type == LST_OTHER))
        {
	    char *content;

	    content = (char *)xmlStrndup (chars, len);
	    if (content == NULL) {
		    /* fatal error */
		    state->ret_code = MEMORY_ERROR;
		    return;
	    }

            /* The first position is default value, and also is CUPS value */
            if ((state->filter_type == 2) && (state->rawfilter == 1) && (atoi (state->id) !=2))
	    {
		char *tmp;
		
		tmp = GetValueCap (content, 1);
		if (tmp == NULL) {
			/* fatal error */
		}
		
                state->content = strdup (tmp);
		FreeValueCap (tmp);
	    }
            else
	    {
		if (state->content == NULL)
		{
		    state->content = strdup (content);
		}
		else
		{
		    char *new;
		    int len;

		    len = strlen (state->content) + strlen (content) + 1;
		    new = (char *) malloc (len);
		    if (new == NULL)
		    {
			/* fatal error */
			state->ret_code = MEMORY_ERROR;
		    }
		    else
		    {
			strcpy (new, state->content);
			strcat (new, content);
			free (state->content);
			state->content = new;
		    }
		}
	    }
	    xmlFree (content);
	    content = NULL;
        }
        break;
    default:
        /* don't care about content in any other states */
        break;
    }
}

static
xmlSAXHandler sax_handler = {
                                (internalSubsetSAXFunc)NULL,                        /* internalSubset */
                                (isStandaloneSAXFunc)NULL,                          /* isStandalone */
                                (hasInternalSubsetSAXFunc)NULL,                     /* hasInternalSubset */
                                (hasExternalSubsetSAXFunc)NULL,                     /* hasExternalSubset */
                                (resolveEntitySAXFunc)NULL,                         /* resolveEntity */
                                (getEntitySAXFunc)NULL,                             /* getEntity */
                                (entityDeclSAXFunc)NULL,                            /* entityDecl */
                                (notationDeclSAXFunc)NULL,                          /* notationDecl */
                                (attributeDeclSAXFunc)NULL,                         /* attributeDecl */
                                (elementDeclSAXFunc)NULL,                           /* elementDecl */
                                (unparsedEntityDeclSAXFunc)NULL,                    /* unparsedEntityDecl */
                                (setDocumentLocatorSAXFunc)NULL,                    /* setDocumentLocator */
                                (startDocumentSAXFunc)pipscap_parser_start_document,/* startDocument */
                                (endDocumentSAXFunc)pipscap_parser_end_document,    /* endDocument */
                                (startElementSAXFunc)pipscap_parser_start_element,  /* startElement */
                                (endElementSAXFunc)pipscap_parser_end_element,      /* endElement */
                                (referenceSAXFunc)NULL,                             /* reference */
                                (charactersSAXFunc)pipscap_parser_characters,       /* characters */
                                (ignorableWhitespaceSAXFunc)NULL,                   /* ignorableWhitespace */
                                (processingInstructionSAXFunc)NULL,                 /* processingInstruction */
                                (commentSAXFunc)NULL,                               /* comment */
                                (warningSAXFunc)NULL,                               /* warning */
                                (errorSAXFunc)NULL,                                 /* error */
                                (fatalErrorSAXFunc)NULL,                            /* fatalError */
                            };


int
pipscap_parser_file(PipsCapParserPtr parser_state, char *xmlfile)
{
    int ret_code = 0;

    if (xmlSAXUserParseFile (&sax_handler, parser_state, xmlfile) < 0)
    {
        if (parser_state->ret_code == 0)
            parser_state->ret_code = NOT_WELL_FORM_ERROR;

        ret_code = parser_state->ret_code;
    }

    if (parser_state->state != PARSER_FINISH)
        ret_code = PARSER_ERROR;

    return ret_code;
}

int
GetDatabase (char *xmlfile, PipsCapParserPtr parser_state)
{
    /* Parsing the specific xmlfile firstly */
    int ret_code = 0;

    parser_state->parser_status = 1;
    if ((ret_code = pipscap_parser_file (parser_state, xmlfile)))
        return ret_code;

    /* Parsing the common epson xmlfile secondly */
    parser_state->parser_status = 2;
    sprintf (xmlfile, "%s", COMMON_DATA_PATH);
    if ((ret_code = CheckExistFile (xmlfile)))
        return ret_code;

    if ((ret_code = pipscap_parser_file (parser_state, xmlfile)))
        return ret_code;

    if (parser_state->filter_type == 2)
    {
        /* Add new quality feature for rawfilter type */
        if (AddCapListCap (&parser_state->data_list->cap_list, P_QUALITY_RAW, "ql"))
            return MEMORY_ERROR;

        PrintModeList curr_mode = parser_state->data_list->mode_color_list;
        while (curr_mode != NULL)
        {
            if (AddElementListCap (&parser_state->data_list->cap_list->list, curr_mode->mode_name, ""))
                return MEMORY_ERROR;
            curr_mode = curr_mode->next;
        }
    }

    return 0;
}


/* Load xml file into tree for small file case */
static int
CheckExistFile (const char* input_file)
{
    if (access (input_file, F_OK | R_OK) == -1)
        return NO_RESOURCE_ERROR;

    return 0;
}

static int
OverCheckXmlFile (const char *rc_file, xmlDocPtr *doc, char* root_name)
{
    xmlNodePtr curr;

    *doc = xmlParseFile (rc_file);

    /* Check well-formed */
    if (*doc == NULL )
        return RESOURCE_FORMAT_ERROR;

    /* Check empty xml file */
    curr = xmlDocGetRootElement(*doc);
    if (curr == NULL)
    {
        xmlFreeDoc(*doc);
        return RESOURCE_EMPTY_ERROR;
    }

    /* Check correct option name of root */
    if (strcmp(curr->name, root_name))
    {
        xmlFreeDoc(*doc);
        return RESOURCE_OPTION_ERROR;
    }

    return 0;
}

int
LoadPipsrcFile (const char *printer_name, char *pipsrc_path, CapContextList *context_list)
{
    xmlDocPtr doc = NULL;
    xmlNodePtr curr;
    char *option, *capid;
    char *temp_name;
    int ret_code = 0, id;

    /* Check exist pipsrc file */
    if ((ret_code = CheckExistFile (pipsrc_path)))
        return PIPSRC_NOT_FOUND;

    if ((ret_code = OverCheckXmlFile (pipsrc_path, &doc, "pipsrc")))
        return ret_code;

    /* Find correct printer */
    curr = xmlDocGetRootElement (doc);

    curr = curr->xmlChildrenNode;
    while (curr != NULL)
    {
        if (!strcmp(curr->name, "printer"))
        {
            if (!(temp_name = xmlGetProp (curr, "name")))
                return RESOURCE_OPTION_ERROR;
            else if (!strcmp (printer_name, temp_name))
            {
                curr = curr->xmlChildrenNode;
                while (curr != NULL)
                {
                    if (!strcmp (curr->name, "option"))
                    {
                        if (!(capid = xmlGetProp (curr, "id")))
                        {
                            ret_code = RESOURCE_OPTION_ERROR;
                            break;
                        }

                        if (!(option = xmlNodeListGetString(doc, curr->xmlChildrenNode, 1)))
                        {
                            ret_code = RESOURCE_VALUES_ERROR;
                            break;
                        }
                        id = atoi (capid);
                        if (AddContextListCap (context_list, id, option))
                        {
                            ret_code = MEMORY_ERROR;
                            break;
                        }

			if (capid != NULL)
			{
			    xmlFree (capid);
			    capid = NULL;
			}
			if (option != NULL)
			{
			    xmlFree (option);
			    option = NULL;
			}
                    }
                    curr = curr->next;
                }
                break;
            }
        }
        curr = curr->next;
    }
    xmlFreeDoc(doc);

    if (!(*context_list))
        return PIPSRC_NOT_FOUND;

    return ret_code;
}

static int
RemovePrinterFromPipsrcFile (const char *printer_name, char *pipsrc_path)
{
    char str_name[MAX_SIZE], *line, *before_buffer, *back_buffer;
    int user_id, file_len;
    long insert_pos, protect_pos;
    FILE *fp;
    size_t len = 0;
    size_t read_size = 0;

    line = before_buffer = back_buffer = 0;
    user_id = geteuid ();

    if ((fp = fopen (pipsrc_path, "rb")) == NULL)
        return FILE_ACCESS_ERROR;

    /* Set name*/
    sprintf (str_name, "%s%s%s%s", "name=", "\"", printer_name, "\"");
    while ((read_size = getline (&line, &len, fp)) != -1)
    {
        if (strstr (line, str_name))
        {
            fseek (fp, read_size * -1, SEEK_CUR);
            break;
        }
    }

    insert_pos = ftell (fp);
    if (insert_pos != 0)
    {
        fseek (fp, 0, SEEK_SET);
        before_buffer = (char *)calloc (insert_pos + 1, sizeof (char));
        fread (before_buffer, sizeof (char), insert_pos, fp);
    }

    while ((read_size = getline (&line, &len, fp)) != -1)
    {
        if (strstr (line, "</printer>"))
            break;
    }

    protect_pos = ftell (fp);
    fseek (fp, 0, SEEK_END);
    file_len = ftell (fp);
    if (file_len != 0)
    {
        back_buffer = (char *)calloc (file_len - protect_pos + 1, sizeof (char));
        fseek (fp, protect_pos, SEEK_SET);
        fread (back_buffer, sizeof (char), file_len, fp);
    }
    fclose (fp);

    if ((fp = fopen (pipsrc_path, "wb")) == NULL)
        return FILE_ACCESS_ERROR;

    if (before_buffer != 0)
    {
        fwrite (before_buffer, sizeof (char), strlen (before_buffer), fp);
        free (before_buffer);
    }

    if (back_buffer != 0)
    {
        fwrite (back_buffer, sizeof (char), strlen(back_buffer), fp);
        free (back_buffer);
    }
    fclose(fp);

    if (!user_id)
        chmod (pipsrc_path, 438);

    return NO_ERROR;
}

static int
CheckExistPrinterFromPipsrcFile (const char *printer_name, char *pipsrc_path)
{
    char *temp_name = NULL;
    int ret_code;
    xmlDocPtr doc = NULL;
    xmlNodePtr curr;

    /* Check exist pipsrc file */
    if (access (pipsrc_path, F_OK))
    {
	/* The pipsrc not insatalled. *cleate* */
	int len;
	char *com;

	len = strlen (PIPS_PRINTER_DIR)
		+ (strlen (printer_name) * 2)
		+ strlen (pipsrc_path)
		+ 20;

	com = (char *) malloc (len);
	if (com == NULL)
	{
	    return MEMORY_ERROR;
	}

	sprintf (com, "cp %s/%s/pipsrc_%s %s", PIPS_PRINTER_DIR, printer_name, printer_name, pipsrc_path);
	system (com);
	free (com);

	chmod(pipsrc_path, 438);
    }

    if ((ret_code = CheckExistFile (pipsrc_path))) {
        return ret_code;
    }

    if ((ret_code = OverCheckXmlFile (pipsrc_path, &doc, "pipsrc")))
        return ret_code;

    /* Find correct printer */
    curr = xmlDocGetRootElement (doc);
    curr = curr->xmlChildrenNode;
    while (curr != NULL)
    {
        if (!strcmp(curr->name, "printer"))
        {
            if (!(temp_name = xmlGetProp (curr, (const xmlChar *)"name")))
                return RESOURCE_OPTION_ERROR;
            else if (!strcmp (printer_name, temp_name))
            {
                ret_code = 0;
                xmlFree (temp_name);
                break;
            }
        }
        curr = curr->next;
    }
    xmlFree (doc);

    return ret_code;
}

int
SavePipsrcFile (const char* printer_name, char *pipsrc_path, CapContextList context_list, int save_type)
{
    int ret_code;
    char capid[MAX_SIZE];
    xmlDocPtr doc = NULL;
    xmlNodePtr root = NULL;

    if (save_type == 1) /* Saving into pipsrc file */
    {
        if ((ret_code = CheckExistPrinterFromPipsrcFile (printer_name, pipsrc_path)))
            return ret_code;
        else
        {   /* Remove printer from pipsrc file */
            if ((ret_code = RemovePrinterFromPipsrcFile (printer_name, pipsrc_path)))
                return ret_code;
        }
        doc = xmlParseFile (pipsrc_path);
        root = xmlDocGetRootElement (doc);
    }

    /* Add new printer name into pipsrc file */
    if (!root )
    {   /* Create new xml document with version 1.0 */
        doc = xmlNewDoc("1.0");
        /* create a new root node "pipsrc" */
        root = xmlNewDocNode(doc, NULL, "pipsrc", NULL);
        xmlDocSetRootElement (doc, root);
        xmlNodeAddContent (root, "\n");
    }
    /* Add <printer> tag */
    xmlNodePtr prt_node = xmlNewNode (NULL, "printer");
    xmlSetProp (prt_node, "name", printer_name);
    xmlNodeAddContent (root, "  ");
    xmlAddChild (root, prt_node);
    xmlNodeAddContent (root, "\n");
    xmlNodeAddContent (prt_node, "\n");
    /* Add <option> tag */
    while (context_list != NULL)
    {
        xmlNodePtr opt_node = xmlNewNode(NULL, "option");

        sprintf (capid, "%d", context_list->capid);
        xmlSetProp (opt_node, "id", (xmlChar *) capid);
        xmlNodeSetContent (opt_node, (xmlChar *) context_list->option);
        xmlNodeAddContent (prt_node, "    ");
        xmlAddChild (prt_node, opt_node);
        xmlNodeAddContent (prt_node, "\n");
        context_list = context_list->next;
    }
    xmlNodeAddContent (prt_node, "  ");
    /* Output */
    xmlSaveFileEnc (pipsrc_path, doc, "UTF-8");
    xmlFreeDoc(doc);

    return NO_ERROR;
}

#if 0
int
main( int argc, char *argv[] )
{
    char *ret_data = NULL;
    int ret_code = 0;
    printf ("Start getting database ..\n");

    PipsCapParser *parser_state = (PipsCapParser*) malloc (sizeof (PipsCapParser));
    parser_state->filter_type = 2;

    ret_code = GetDatabase (argv[1], parser_state);
    if (ret_code != 0)
    {
        ret_data = strdup (parser_state->ret_data);
        return ret_code;
    }

    printf ("Finish getting database ..\n");

    return 0;
}
#endif
