/*
 * Photo Image Print System
 * Copyright (C) 2000-2005 EPSON AVASYS Corporation.
 * Copyright (C) Seiko Epson Corporation 2000-2008.
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, EPSON AVASYS Corporation gives permission to
 * link the code of this program with libraries which are covered by
 * the EPSON AVASYS Public License and distribute their linked
 * combinations.  You must obey the GNU General Public License in all
 * respects for all of the code used other than the libraries which
 * are covered by EPSON AVASYS Public License.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <libxml/xmlmemory.h>
#include "pipserr.h"
#include "pipsmem.h"
#include "setup.h"

static void FreeTagValue (char*);
static char* GetTagValue (xmlDocPtr, const xmlNodePtr, char*);

int
CheckExistFile(char* input_file)
{
    if (access (input_file, F_OK | R_OK) == -1)
    {
        FILTER_ERROR_HANDLER (NO_RESOURCE_ERROR, input_file);
        return 1;
    }
    return 0;
}

int
OverCheckXmlFile (char *rc_file, xmlDocPtr *doc, xmlChar *root_name)
{
    xmlNodePtr cur;

    if (CheckExistFile (rc_file))
        return 1;

    *doc = xmlParseFile (rc_file);
    /* Check well-formed */
    if (*doc == NULL )
    {
        FILTER_ERROR_HANDLER (RESOURCE_FORMAT_ERROR, rc_file);
        return 1;
    }
    /* Check empty xml file */
    cur = xmlDocGetRootElement(*doc);
    if (cur == NULL)
    {
        FILTER_ERROR_HANDLER (RESOURCE_EMPTY_ERROR, rc_file);
        xmlFreeDoc(*doc);
        return 1;
    }
    /* Check correct option name of root */
    if (strcmp(cur->name, root_name))
    {
        FILTER_ERROR_HANDLER (RESOURCE_OPTION_ERROR, root_name);
        xmlFreeDoc(*doc);
        return 1;
    }

    return 0;
}

static void
FreeTagValue (char *value)
{
    if (value != NULL) {
	free (value);
	value = NULL;
    }
    
    return;
}

static char*
GetTagValue (const xmlDocPtr doc, const xmlNodePtr parent, char *tag_name)
{
    xmlChar *value = NULL;
    xmlNodePtr cur;

    /* Find all child of this current tag */
    cur = parent->xmlChildrenNode;

    while (cur != NULL)
    {
        if (!strcmp((char*)cur->name, tag_name))
        {
	    char *tmp;

            if (!(tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1)))
                FILTER_ERROR_HANDLER (RESOURCE_VALUES_ERROR, tag_name);
	    
	    value = strdup (tmp);
	    if (value == NULL) {
		    /* error */
	    }

	    xmlFree (tmp);
            break;
        }
	
        /* Go to the next child */
        cur = cur->next;
    }
    if (cur == NULL)
        FILTER_ERROR_HANDLER (RESOURCE_OPTION_ERROR, tag_name);

    return (char *)value;
}


int
CheckPrinterName(char* alias_name, short used_type)
{
    int ret_code = 1;
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        xmlChar *value = NULL;
        char rc_file[] = PIPS_EKPD_RC_PATH;

        if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
            return 1;

        cur = xmlDocGetRootElement(doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if (!strcmp(cur->name, "printer"))
            {
                if ((value = GetTagValue (doc, cur, "aliasname")))
                {
                    if (!strcmp (value, alias_name))
                    {
                        ret_code = 0;
                        break;
                    }
                } /* Value error */
            }
            cur = cur->next;
        }
        if (value != NULL)
	{
            FreeTagValue (value);
	    value = NULL;
	}
        xmlFreeDoc(doc);

    if (ret_code)
        FILTER_ERROR_HANDLER (PRINTER_NOT_FOUND, alias_name);

    return ret_code;
}

char*
GetDefaultPrinter (short used_type)
{
    char *model_name = NULL ;
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        char *temp_name = NULL, rc_file[] = PIPS_EKPD_RC_PATH;

        if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
            return NULL;

        cur = xmlDocGetRootElement(doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if (!strcmp(cur->name, "defaultprinter"))
            {
                if (!(temp_name = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1)))
                    FILTER_ERROR_HANDLER (RESOURCE_VALUES_ERROR, "defaultprinter");
                else
                {
                    model_name = (char *) malloc (strlen (temp_name) + 1);
                    strcpy (model_name, temp_name);
		    xmlFree (temp_name);
		    temp_name = NULL;
                }
                break;
            }
            cur = cur->next;
        }

        xmlFreeDoc(doc);

    return model_name;

}

char*
GetAliasName(char *model_name)
{
	char *alias_name = NULL;
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        xmlChar *value;
        char rc_file[] = PIPS_EKPD_RC_PATH;
        char *temp_name = NULL;
	
        if (OverCheckXmlFile(rc_file, &doc, "ekpdrc")) {
		return NULL;
	}

        cur = xmlDocGetRootElement(doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL) {
		if (!strcmp(cur->name, "printer")) {
			if (!(temp_name = xmlGetProp(cur, (const xmlChar *)"name"))) {
				FILTER_ERROR_HANDLER (RESOURCE_ATTRIBUTE_ERROR, "<printer name=");
			}
			if (!strcmp(model_name, temp_name)) {
				if ((value = GetTagValue(doc, cur, "aliasname"))) {
					alias_name = strdup(value);
					xmlFree(temp_name);
					FreeTagValue(value);
					temp_name = value = NULL;
					break;
				}
			}
			if (temp_name != NULL) {
				xmlFree(temp_name);
				temp_name = NULL;
			}
		}
		cur = cur->next;
        }
        xmlFreeDoc(doc);
	
	return alias_name;
}

char*
GetModelName (char *alias_name, short used_type)
{
    char *model_name = NULL;
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        xmlChar *value;
        char rc_file[] = PIPS_EKPD_RC_PATH;
        char *temp_name = NULL;

        if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
            return NULL;

        cur = xmlDocGetRootElement (doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if (!strcmp (cur->name, "printer"))
            {
                if ((value = GetTagValue (doc, cur, "aliasname")))
                {
                    if (!strcmp (value, alias_name))
                    {
                        if (!(temp_name = xmlGetProp (cur, (const xmlChar *)"name")))
                            FILTER_ERROR_HANDLER (RESOURCE_ATTRIBUTE_ERROR, "<printer name=");
                        else
                            model_name = strdup (temp_name);
                        xmlFree (temp_name);
                        FreeTagValue (value);
			temp_name = value = NULL;
                        break;
                    }
                }
		if (value != NULL)
		{
		    FreeTagValue (value);
		    value = NULL;
		}
            }
            cur = cur->next;
        }
        xmlFreeDoc(doc);

    return model_name;
}

char**
GetModelListName (short used_type)
{
    char** model_list=(char**)malloc(128*sizeof(char*));
    int count = 0;

    if (used_type == PNG_TYPE)
    {
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        char rc_file[] = PIPS_EKPD_RC_PATH;
        char* model_name;

        if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
            return NULL;


        cur = xmlDocGetRootElement (doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if (!xmlStrcmp (cur->name, (const xmlChar *)"printer"))
            {
                if (!(model_name = xmlGetProp (cur, (const xmlChar *)"name")))
                {
                    FILTER_ERROR_HANDLER (RESOURCE_ATTRIBUTE_ERROR, "<printer name=");
                    break;
                }
                else
                {
                    model_list[count]=(char*)malloc(strlen(model_name)+1);
                    strcpy(model_list[count++],model_name);
                }

            }
            cur = cur->next;
        }
        xmlFreeDoc(doc);
    }
    model_list[count]=NULL;
    return model_list;
};

char**
GetSubListName (char *sub_name)
{
    char** alias_list=(char**)malloc(128*sizeof(char*));
    int count = 0;

    xmlDocPtr doc = NULL;
    xmlNodePtr cur;
    xmlChar *alias_name;
    char rc_file[] = PIPS_EKPD_RC_PATH;
    if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
        return NULL;
    cur = xmlDocGetRootElement (doc);
    /* Go to the next child */
    cur = cur->xmlChildrenNode;
    while (cur != NULL)
    {
        if (!xmlStrcmp (cur->name, (const xmlChar *)"printer"))
        {
            xmlNodePtr tmp;
            tmp = cur->xmlChildrenNode;
            while (tmp != NULL)
            {
                if ((!xmlStrcmp(tmp->name, (const xmlChar *)sub_name)))
                {
                    alias_name = xmlNodeListGetString(doc, tmp->xmlChildrenNode, 1);
                    alias_list[count]=(char*)malloc(strlen(alias_name)+1);
                    strcpy(alias_list[count++],alias_name);
                }
                tmp = tmp->next;
            }
        }
        cur = cur->next;
    }
    xmlFreeDoc(doc);

    alias_list[count]=NULL;
    return alias_list;
};

char*
GetLibPrtPath (char* model_name)
{
    char lib_path[MAX_SIZE];

    memset (lib_path, '\0', MAX_SIZE);
    sprintf (lib_path, "%s%s%s%s", "/usr/lib","/lib", model_name, ".so");

    if (CheckExistFile (lib_path))
        return NULL;

    return strdup (lib_path);
}

char *
GetCapValue (const CapContextList p_caplist, int capid)
{
    CapContextList curr = p_caplist;
    while (curr != NULL)
    {
        if (curr->capid == capid)
            return curr->option;
        else
            curr = curr->next;
    }

    return NULL;
}

char *
GetOptValue (const ElementList list, char *option)
{
    ElementList curr = list;
    while (list)
    {
        if (!strcmp (curr->option, option))
            return curr->value;
        else
            curr = curr->next;
    }

    return NULL;
}


int
UpdateContextList (CapContextList p_caplist, int capid, char *option)
{
    CapContextList curr = p_caplist;
    while (curr)
    {
        if (curr->capid == capid)
        {
            free (curr->option);
            curr->option = strdup (option);
            return NO_ERROR;
        }
        else
            curr = curr->next;
    }

    return 1;
}

int
AddContextList (CapContextList *p_caplist, int capid, char *option)
{
    CapContextList node = (CapContextList) malloc (sizeof (CapContextNode));
    if (!node)
        return 1;

    node->capid = capid ;
    node->option = strdup (option);
    node->next = *p_caplist;
    *p_caplist = node;

    return NO_ERROR;
}

int
AddPaperRsc (CapContextList *p_caplist, LP_PS_STRUCT paper_conf, BOOL add_type)
{
    char value[MAX_SIZE];

    sprintf (value, "%ld", paper_conf->size.x);
    if (AddContextList (p_caplist, P_PAPER_SIZE_X, value))
        return 1;
    sprintf (value, "%ld", paper_conf->size.y);
    if (AddContextList (p_caplist, P_PAPER_SIZE_Y, value))
        return 1;
    sprintf (value, "%ld", paper_conf->area.x);
    if (AddContextList (p_caplist, P_PAPER_AREA_X, value))
        return 1;
    sprintf (value, "%ld", paper_conf->area.y);
    if (AddContextList (p_caplist, P_PAPER_AREA_Y, value))
        return 1;
    sprintf (value, "%ld", paper_conf->def_area.x);
    if (AddContextList (p_caplist, P_DEF_PAPER_AREA_X, value))
        return 1;
    sprintf (value, "%ld", paper_conf->def_area.y);
    if (AddContextList (p_caplist, P_DEF_PAPER_AREA_Y, value))
        return 1;
    sprintf (value, "%ld", paper_conf->margin.x);
    if (add_type)
    {
        if (UpdateContextList (*p_caplist, P_MARGIN_X, value ))
            return 1;
    }
    else
    {
        if (AddContextList (p_caplist, P_MARGIN_X, value))
            return 1;
    }
    sprintf (value, "%ld", paper_conf->margin.y);
    if (add_type)
    {
        if (UpdateContextList (*p_caplist, P_MARGIN_Y, value ))
            return 1;
    }
    else
    {
        if (AddContextList (p_caplist, P_MARGIN_Y, value))
            return 1;
    }

    if (add_type)
    {
        sprintf (value, "%ld", paper_conf->def_margin.x);
        if (AddContextList (p_caplist, P_DEF_MARGIN_X, value))
            return 1;
        sprintf (value, "%ld", paper_conf->def_margin.y);
        if (AddContextList (p_caplist, P_DEF_MARGIN_Y, value))
            return 1;
    }

    return NO_ERROR;
}

char*
GetDefaultAliasName (short used_type)
{
    char *alias_name = NULL ;
    if (used_type == PNG_TYPE)
    {
        xmlDocPtr doc = NULL;
        xmlNodePtr cur;
        char *def_prt_name, *temp_prt_name, *temp_alias_name, rc_file[] = PIPS_EKPD_RC_PATH;

        def_prt_name = temp_prt_name =  temp_alias_name = NULL;
        if (OverCheckXmlFile (rc_file, &doc, "ekpdrc"))
            return NULL;

        cur = xmlDocGetRootElement(doc);
        /* Go to the next child */
        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if (!def_prt_name && !strcmp(cur->name, "defaultprinter"))
            {
                if (!(def_prt_name = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1)))
                {
                    FILTER_ERROR_HANDLER (RESOURCE_VALUES_ERROR, "defaultprinter");
                    break;
                }
            }

            if (def_prt_name && !strcmp(cur->name, "printer"))
            {
                if (!(temp_prt_name = xmlGetProp (cur, (const xmlChar *)"name")))
                {
                    FILTER_ERROR_HANDLER (RESOURCE_ATTRIBUTE_ERROR, "<printer name=");
                    break;
                }
                else
                {
                    if (!strcmp (def_prt_name, temp_prt_name))
                    {
                        if ((temp_alias_name = GetTagValue (doc, cur, "aliasname")))
			{
                            alias_name = strdup (temp_alias_name);
			    FreeTagValue (temp_alias_name);
			    temp_alias_name = NULL;
			}
                        break;
                    }
                }

		if (temp_prt_name)
		{
		    xmlFree (temp_prt_name);
		    temp_prt_name = NULL;
		}
            }
            cur = cur->next;
        }
        if (def_prt_name)
            xmlFree (def_prt_name);
        if (temp_prt_name)
            xmlFree (temp_prt_name);
        if (temp_alias_name)
            free (temp_alias_name);

        xmlFreeDoc (doc);
    }
    else /* if (used_type == RAW_TYPE) */
    {}

    return alias_name;

}


char *
str_clone (char *str, size_t size)
{
    char *new;

    if (!str || !size)
        return NULL;    /* error */

    new = pips_malloc (char, size + 1);

    strncpy (new, str, size);
    new[size] = '\0';

    return new;
}

long
Round_Off (double val)
{
    long low, high;

    low = (long)val;
    high = (long)(val + 0.5);

    return (low == high) ? low : high;
}

SpoolerType
GetActiveSpooler(void)
{
	int status;
	int cups_result;
	int lpr_result;

	/* search process. */
	status = system("/sbin/pidof cupsd 2>&1 >/dev/null");
	cups_result = WEXITSTATUS(status);
	status = system("/sbin/pidof lpd 2>&1 >/dev/null");
	lpr_result = WEXITSTATUS(status);

	if (cups_result == 0 && lpr_result == 0) {
		return SPL_Conflict;

	} else if (cups_result == 0) {
		return SPL_CUPS;
		
	} else if (lpr_result == 0) {
		return SPL_LPRng;

	}
       
	return SPL_NoSupported;
}
